/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>

#include "config.h"

#include "mars/core.h"
#include "mars/error.h"
#include "mars/mutex.h"
#include "mars/workload_queue.h"

#include "context_internal.h"
#include "kernel_internal_types.h"
#include "workload_internal_types.h"

static inline uint64_t workload_queue_block_ea(uint64_t queue_ea, int block)
{
	return queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;
}

static inline uint64_t workload_queue_workload_ea(uint64_t queue_ea,
						  int workload_id)
{
	uint64_t context_ea =
		mars_ea_get_uint64(
			queue_ea +
			offsetof(struct mars_workload_queue_header,
				 context_ea));
	return context_ea +
		sizeof(struct mars_workload_context) * workload_id;
}

static inline uint64_t workload_queue_block_bits_ea(uint64_t block_ea,
						    int index)
{
	return block_ea +
		offsetof(struct mars_workload_queue_block, bits) +
		sizeof(uint64_t) * index;
}

int mars_workload_queue_create(struct mars_context *mars)
{
	struct mars_workload_queue *queue;
	int block;
	int index;
	uint64_t queue_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (mars->workload_queue_ea)
		return MARS_ERROR_STATE;

	/* allocate workload instance */
	queue_ea = mars_ea_memalign(
		MARS_WORKLOAD_QUEUE_ALIGN,
		sizeof(struct mars_workload_queue));
	if (!queue_ea)
		return MARS_ERROR_MEMORY;

	/* prepare work area for queue header */
	queue = mars_ea_work_area_get(queue_ea,
		MARS_WORKLOAD_QUEUE_ALIGN,
		sizeof(struct mars_workload_queue_header));

	/* initialize workload queue header */
	queue->header.flag = MARS_WORKLOAD_QUEUE_FLAG_NONE;
	queue->header.queue_ea = queue_ea;
	queue->header.context_ea =
		queue_ea + offsetof(struct mars_workload_queue, context);

	/* update queue header on EA */
	mars_ea_put(queue_ea, queue, sizeof(struct mars_workload_queue_header));

	/* initialize workload queue blocks */
	for (block = 0; block < MARS_WORKLOAD_NUM_BLOCKS; block++) {
		uint64_t block_ea = workload_queue_block_ea(queue_ea, block);

		for (index = 0; index < MARS_WORKLOAD_PER_BLOCK; index++) {
			uint64_t bits_ea =
				workload_queue_block_bits_ea(block_ea, index);
			uint64_t bits = 0;

			MARS_BITS_SET(&bits, STATE,
				MARS_WORKLOAD_STATE_NONE);
			MARS_BITS_SET(&bits, PRIORITY,
				MARS_WORKLOAD_PRIORITY_MIN);
			MARS_BITS_SET(&bits, COUNTER,
				MARS_WORKLOAD_COUNTER_MIN);
			MARS_BITS_SET(&bits, SIGNAL,
				MARS_WORKLOAD_SIGNAL_OFF);
			MARS_BITS_SET(&bits, WAIT_ID,
				MARS_WORKLOAD_ID_NONE);

			mars_ea_put_uint64(bits_ea, bits);
		}

		mars_mutex_reset(block_ea);
	}

	/* sync EA */
	mars_ea_sync();

	/* set the workload queue instance in the mars context */
	mars->workload_queue_ea = queue_ea;

	return MARS_SUCCESS;
}

int mars_workload_queue_destroy(struct mars_context *mars)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint16_t id = 0;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* check for any workloads left in workload queue */
	while (id < MARS_WORKLOAD_MAX) {
		uint64_t block_ea, bits_ea;
		uint64_t bits;

		block = id / MARS_WORKLOAD_PER_BLOCK;
		index = id % MARS_WORKLOAD_PER_BLOCK;

		/* get bits from EA */
		block_ea = workload_queue_block_ea(queue_ea, block);
		bits_ea = workload_queue_block_bits_ea(block_ea, index);
		bits = mars_ea_get_uint64(bits_ea);

		if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_NONE)
			return MARS_ERROR_STATE;

		id++;
	}

	/* free workload queue instance */
	mars_ea_free(queue_ea);

	/* set the workload queue to NULL for error checking */
	mars->workload_queue_ea = 0;

	return MARS_SUCCESS;
}

int mars_workload_queue_exit(struct mars_context *mars)
{
	uint64_t queue_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	mars_ea_put_uint32(queue_ea +
			   offsetof(struct mars_workload_queue_header, flag),
			   MARS_WORKLOAD_QUEUE_FLAG_EXIT);

	return MARS_SUCCESS;
}

int mars_workload_queue_add_begin(struct mars_context *mars,
				uint16_t *id,
				uint64_t *workload_ea)
{
	uint64_t queue_ea;
	int block = 0;
	int index = 0;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (!id)
		return MARS_ERROR_NULL;

	queue_ea = mars->workload_queue_ea;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* initialize return id to 0 */
	*id = 0;

	while (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_NONE) {
		(*id)++;
		index++;
		if (index == MARS_WORKLOAD_PER_BLOCK) {
			index = 0;

			mars_mutex_unlock(block_ea);

			if (++block == MARS_WORKLOAD_NUM_BLOCKS)
				return MARS_ERROR_LIMIT;

			block_ea = workload_queue_block_ea(queue_ea, block);
			mars_mutex_lock(block_ea);
		}
		bits_ea = workload_queue_block_bits_ea(block_ea, index);
		bits = mars_ea_get_uint64(bits_ea);
	}

	/* set state to adding */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_ADDING);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	/* if requested set workload context pointer to return */
	if (workload_ea)
		*workload_ea = workload_queue_workload_ea(queue_ea, *id);

	return MARS_SUCCESS;
}

int mars_workload_queue_add_end(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_ADDING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* reset workload queue bits and set state to finished state */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_FINISHED);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

int mars_workload_queue_add_cancel(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_ADDING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state back to none state */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_NONE);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_begin(struct mars_context *mars,
				uint16_t id,
				uint64_t *workload_ea)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_FINISHED) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state to removing */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_REMOVING);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	/* if requested set workload context pointer to return */
	if (workload_ea)
		*workload_ea = workload_queue_workload_ea(queue_ea, id);

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_end(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_REMOVING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state to none */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_NONE);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_cancel(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_REMOVING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state back to finished */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_FINISHED);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_begin(struct mars_context *mars,
				uint16_t id, uint8_t priority,
				uint64_t *workload_ea)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_FINISHED) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* reset workload queue bits and set state to scheduling */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_SCHEDULING);
	MARS_BITS_SET(&bits, PRIORITY, priority);
	MARS_BITS_SET(&bits, COUNTER, MARS_WORKLOAD_COUNTER_MIN);
	MARS_BITS_SET(&bits, SIGNAL, MARS_WORKLOAD_SIGNAL_OFF);
	MARS_BITS_SET(&bits, WAIT_ID, MARS_WORKLOAD_ID_NONE);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	/* if requested set workload context pointer to return */
	if (workload_ea)
		*workload_ea = workload_queue_workload_ea(queue_ea, id);

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_end(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_SCHEDULING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state to ready */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_READY);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_cancel(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_SCHEDULING) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set state back to finished */
	MARS_BITS_SET(&bits, STATE, MARS_WORKLOAD_STATE_FINISHED);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}

static int test_workload_state_finished(uint32_t upper, void *param)
{
	/* this function assumes 'STATE' is stored in upper 32bits */
	uint64_t bits = (uint64_t)upper << 32;

	(void)param;

	switch (MARS_BITS_GET(&bits, STATE)) {
	case MARS_WORKLOAD_STATE_FINISHED:
		return MARS_SUCCESS;
	case MARS_WORKLOAD_STATE_NONE:
		return MARS_ERROR_STATE;
	default:
		return -1;
	}
}

int mars_workload_queue_wait(struct mars_context *mars,
				uint16_t id,
				uint64_t *workload_ea)
{
	int ret;
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	ret = mars_ea_cond_wait(bits_ea, test_workload_state_finished, NULL);
	if (ret)
		return ret;

	/* if requested set workload context pointer to return */
	if (workload_ea)
		*workload_ea = workload_queue_workload_ea(queue_ea, id);

	return MARS_SUCCESS;
}

int mars_workload_queue_try_wait(struct mars_context *mars,
				uint16_t id,
				uint64_t *workload_ea)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) == MARS_WORKLOAD_STATE_NONE)
		return MARS_ERROR_STATE;

	/* check if workload is finished */
	if (MARS_BITS_GET(&bits, STATE) != MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	/* if requested set workload context pointer to return */
	if (workload_ea)
		*workload_ea = workload_queue_workload_ea(queue_ea, id);

	return MARS_SUCCESS;
}

int mars_workload_queue_signal_send(struct mars_context *mars,
				uint16_t id)
{
	uint64_t queue_ea;
	int block;
	int index;
	uint64_t bits;
	uint64_t block_ea, bits_ea;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!mars->workload_queue_ea)
		return MARS_ERROR_PARAMS;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	queue_ea = mars->workload_queue_ea;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* prepare work area for queue block */
	block_ea = workload_queue_block_ea(queue_ea, block);

	mars_mutex_lock(block_ea);

	/* get bits from workload queue block */
	bits_ea = workload_queue_block_bits_ea(block_ea, index);
	bits = mars_ea_get_uint64(bits_ea);

	/* check for valid state */
	if (MARS_BITS_GET(&bits, STATE) == MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock(block_ea);
		return MARS_ERROR_STATE;
	}

	/* set signal bit on */
	MARS_BITS_SET(&bits, SIGNAL, MARS_WORKLOAD_SIGNAL_ON);

	mars_ea_put_uint64(bits_ea, bits);

	mars_mutex_unlock(block_ea);

	return MARS_SUCCESS;
}
