/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <spu_mfcio.h>

#include "config.h"

#include "mars/dma.h"
#include "mars/error.h"
#include "mars/module.h"
#include "mars/mutex.h"

#include "kernel_internal_types.h"
#include "workload_internal_types.h"

#define MARS_WORKLOAD_RESERVED_NONE	0
#define MARS_WORKLOAD_RESERVED		1

/* kernel */
void *_kernel_stack;
static struct mars_kernel_params kernel_params;

/* workload queue */
static struct mars_workload_queue_header queue_header;
static struct mars_workload_queue_block queue_block;

/* workload */
static struct mars_workload_context workload;
static uint8_t workload_state;
static uint16_t workload_id;
static uint64_t workload_ea;

/* module entry */
typedef void (*module_entry)(
	const struct mars_kernel_syscalls *kernel_syscalls);

static void module_exit(uint8_t state)
{
	workload_state = state;

	/* restore kernel stack pointer and goto to exit label */
	asm volatile (
		"lqa	$sp, _kernel_stack;"
		"br	_module_exit;"
	);
}

static uint32_t get_ticks(void)
{
	return kernel_params.kernel_ticks.offset - spu_read_decrementer();
}

static uint64_t get_mars_context_ea(void)
{
	return kernel_params.mars_context_ea;
}

static uint32_t get_kernel_id(void)
{
	return kernel_params.kernel_id;
}

static uint16_t get_workload_id(void)
{
	return workload_id;
}

static struct mars_workload_context *get_workload(void)
{
	return &workload;
}

static struct mars_workload_queue_header *get_workload_queue_header(void)
{
	return &queue_header;
}

static struct mars_workload_queue_block *get_workload_queue_block(void)
{
	return &queue_block;
}

static struct mars_kernel_syscalls kernel_syscalls =
{
	module_exit,
	get_ticks,
	get_mars_context_ea,
	get_kernel_id,
	get_workload_id,
	get_workload,
	get_workload_queue_header,
	get_workload_queue_block
};

static void get_block(int block, struct mars_workload_queue_block *dst)
{
	mars_dma_get_and_wait(dst,
		queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block,
		sizeof(struct mars_workload_queue_block),
		MARS_DMA_TAG);
}

static int search_block(int block)
{
	int i;
	int index = -1;
	uint16_t max_counter = 0;
	uint16_t max_priority = 0;

	/* search through workload queue for next workload to run
	 * while incrementing wait counter for all waiting workloads
	 * and pick the workload that has been waiting the longest
	 */
	for (i = 0; i < MARS_WORKLOAD_PER_BLOCK; i++) {
		uint64_t *bits = &queue_block.bits[i];
		uint8_t signal = MARS_BITS_GET(bits, SIGNAL);
		uint8_t priority = MARS_BITS_GET(bits, PRIORITY);
		uint16_t wait_id = MARS_BITS_GET(bits, WAIT_ID);
		uint16_t counter = MARS_BITS_GET(bits, COUNTER);

		switch (MARS_BITS_GET(bits, STATE)) {
		case MARS_WORKLOAD_STATE_READY:
			/* compare priority and counter with previous ones */
			if (index < 0 || priority > max_priority ||
			  (priority == max_priority && counter > max_counter)) {
				index = i;
				max_counter = counter;
				max_priority = priority;
			}
			/* increment wait counter without overflowing */
			if (counter < MARS_WORKLOAD_COUNTER_MAX)
				MARS_BITS_SET(bits, COUNTER, counter + 1);
			break;
		case MARS_WORKLOAD_STATE_WAITING:
			/* waiting for workload to finish so check status */
			if (wait_id != MARS_WORKLOAD_ID_NONE) {
				struct mars_workload_queue_block wait_block;
				struct mars_workload_queue_block *p_wait_block;
				uint8_t wait_state;

				int bl = wait_id / MARS_WORKLOAD_PER_BLOCK;
				int id = wait_id % MARS_WORKLOAD_PER_BLOCK;

				/* check if workload id is in the same block */
				if (block != bl) {
					/* fetch the necessary block */
					get_block(bl, &wait_block);
					/* set pointer to check fetched block */
					p_wait_block = &wait_block;
				} else {
					/* set pointer to check current block */
					p_wait_block = &queue_block;
				}

				wait_state =
					MARS_BITS_GET(&p_wait_block->bits[id],
					STATE);

				/* check if workload is finished and reset */
				if (wait_state ==
					MARS_WORKLOAD_STATE_FINISHED) {
					MARS_BITS_SET(bits, WAIT_ID,
						MARS_WORKLOAD_ID_NONE);
					MARS_BITS_SET(bits, STATE,
						MARS_WORKLOAD_STATE_READY);
				}
			/* waiting for signal so check signal bit and reset */
			} else if (signal == MARS_WORKLOAD_SIGNAL_ON) {
				MARS_BITS_SET(bits, SIGNAL,
					MARS_WORKLOAD_SIGNAL_OFF);
				MARS_BITS_SET(bits, STATE,
					MARS_WORKLOAD_STATE_READY);
				i--;
			}
			break;
		default:
			break;
		}
	}

	/* returns -1 if no runnable workload found */
	return index;
}

static int reserve_block(int block)
{
	int index;
	uint64_t block_ea;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* set the workload index */
	index = search_block(block);
	if (index >= 0) {
		/* update the current state of the workload */
		MARS_BITS_SET(&queue_block.bits[index], STATE,
			MARS_WORKLOAD_STATE_RUNNING);

		/* reset the counter for reserved workload */
		MARS_BITS_SET(&queue_block.bits[index], COUNTER,
			MARS_WORKLOAD_COUNTER_MIN);
	}

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return index;
}

static void release_block(int block, int index)
{
	uint64_t block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* update current workload state in workload queue block */
	MARS_BITS_SET(&queue_block.bits[index], STATE, workload_state);

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);
}

static int __attribute__((noinline)) reserve_workload(void)
{
	int block;
	int index = -1;

	/* search workload queue blocks until runnable workload reserved */
	for (block = 0; block < MARS_WORKLOAD_NUM_BLOCKS; block++) {
		index = reserve_block(block);
		if (index >= 0)
			break;
	}

	/* no runnable workload found */
	if (index < 0)
		return MARS_WORKLOAD_RESERVED_NONE;

	/* set global workload info based on workload block and index */
	workload_id = MARS_WORKLOAD_PER_BLOCK * block + index;
	workload_ea = queue_header.context_ea +
		workload_id * sizeof(struct mars_workload_context);

	/* get the workload context code from workload queue */
	mars_dma_get_and_wait(&workload, workload_ea,
		sizeof(struct mars_workload_context), MARS_DMA_TAG);

	return MARS_WORKLOAD_RESERVED;
}

static void __attribute__((noinline)) release_workload(void)
{
	int block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_id % MARS_WORKLOAD_PER_BLOCK;

	/* put the workload context into workload queue */
	mars_dma_put_and_wait(&workload, workload_ea,
		sizeof(struct mars_workload_context), MARS_DMA_TAG);

	/* release block reservation */
	release_block(block, index);
}

static int __attribute__((noinline)) scheduler(void)
{
	int status;

	/* get the workload queue header */
	mars_dma_get_and_wait(&queue_header,
		kernel_params.workload_queue_ea +
		offsetof(struct mars_workload_queue, header),
		sizeof(struct mars_workload_queue_header),
		MARS_DMA_TAG);

	/* return exit status if exit flag is set from host */
	if (queue_header.flag == MARS_WORKLOAD_QUEUE_FLAG_EXIT)
		return MARS_KERNEL_STATUS_EXIT;

	/* reserve next workload to run or return idle status if none found */
	status = reserve_workload();

	/* return idle status if no workload was reserved */
	if (status == MARS_WORKLOAD_RESERVED_NONE)
		return MARS_KERNEL_STATUS_IDLE;

	/* load the exec code into mpu storage from host storage */
	mars_dma_get_and_wait((void *)MARS_WORKLOAD_MODULE_BASE_ADDR,
		workload.module.exec_ea, workload.module.exec_size,
		MARS_DMA_TAG);

	/* 0 the bss section */
	memset((void *)MARS_WORKLOAD_MODULE_BASE_ADDR +
		workload.module.exec_size, 0, workload.module.bss_size);

	/* save kernel stack pointer */
	asm volatile (
		"stqa	$sp, _kernel_stack;"
	);

	/* sync before executing loaded code */
	spu_sync();

	/* call module entry function */
	((module_entry)workload.module.entry)(&kernel_syscalls);

	/* label so module can jump back here at exit */
	asm volatile (
		"_module_exit:"
	);

	/* release reservation of current workload */
	release_workload();

	return MARS_KERNEL_STATUS_BUSY;
}

static void get_params(uint64_t kernel_params_ea)
{
	int mask;

	/* save event mask */
	mask = spu_read_event_mask();

	/* set event mask for the lost event */
	spu_write_event_mask(MFC_LLR_LOST_EVENT);

	/* set sync begin flag so host knows to begin sync process */
	do {
		mfc_getllar(&kernel_params, kernel_params_ea, 0, 0);
		mfc_read_atomic_status();

		/* set the flag */
		kernel_params.kernel_ticks.flag =
			MARS_KERNEL_TICKS_FLAG_SYNC_BEGIN;
		spu_dsync();

		mfc_putllc(&kernel_params, kernel_params_ea, 0, 0);
	} while (mfc_read_atomic_status() & MFC_PUTLLC_STATUS);

	/* get the tick offset from host and check if sync end flag is set */
	do {
		spu_write_event_ack(MFC_LLR_LOST_EVENT);

		mfc_getllar(&kernel_params, kernel_params_ea, 0, 0);
		mfc_read_atomic_status();

		/* host set the sync end flag so finish */
		if (kernel_params.kernel_ticks.flag ==
			MARS_KERNEL_TICKS_FLAG_SYNC_END)
			break;

		spu_read_event_status();
	} while (1);

	/* now, kernel parameters, including offset of tick counters,
	 * are stored in 'kernel_params'.
	 */

	/* start decrementer */
	spu_write_decrementer(0);

	/* restore event mask */
	spu_write_event_mask(mask);

	/* clear any remnant lost event */
	spu_write_event_ack(MFC_LLR_LOST_EVENT);
}

int main(unsigned long long mpu_context_id,
	unsigned long long kernel_params_ea)
{
	(void)mpu_context_id;

	int exit_flag = 0;

	get_params(kernel_params_ea);

	while (!exit_flag) {
		int status = scheduler();

		switch (status) {
		case MARS_KERNEL_STATUS_BUSY:
			break;
		case MARS_KERNEL_STATUS_IDLE:
			break;
		case MARS_KERNEL_STATUS_EXIT:
			exit_flag = 1;
			break;
		}
	}

	return MARS_SUCCESS;
}
