/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <elf.h>
#include <libspe2.h>
#include <string.h>

#include "config.h"

#include <mars/context.h>
#include <mars/core.h>
#include <mars/error.h>
#include <mars/workload_queue.h>

#include "mars/task.h"

#include "task_internal_types.h"

extern const unsigned char mars_task_module_entry[];

const struct mars_task_context_save_unit mars_task_context_save_all[] =
{
	{ .addr = 0, .size = MARS_TASK_CONTEXT_SAVE_SIZE_MAX },
	{ .addr = 0, .size = 0 },
};

static uint64_t task_exit_code_ea(uint64_t task_ea)
{
	return task_ea + offsetof(struct mars_task_context, exit_code);
}

int mars_task_create(struct mars_context *mars,
		struct mars_task_id *id_ret,
		const char *name, const void *elf_image,
		const struct mars_task_context_save_unit *context_save_unit)
{
	int ret;

	Elf32_Ehdr *ehdr_module, *ehdr_task;
	Elf32_Phdr *phdr_module, *phdr_task;

	uint16_t workload_id;
	uint64_t workload_ea;
	struct mars_task_context *task;

	const struct mars_task_context_save_unit *p = context_save_unit;
	int context_save_area_size = 0;
	int context_save_unit_count = 1;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!id_ret)
		return MARS_ERROR_NULL;
	if (!elf_image)
		return MARS_ERROR_NULL;
	if (name && strlen(name) > MARS_TASK_NAME_LEN_MAX)
		return MARS_ERROR_PARAMS;

	/* process elf header information */
	ehdr_module = (Elf32_Ehdr *)mars_task_module_entry;
	phdr_module = (Elf32_Phdr *)
		((void *)ehdr_module + ehdr_module->e_phoff);

	ehdr_task = (Elf32_Ehdr *)elf_image;
	phdr_task = (Elf32_Phdr *)
		((void *)ehdr_task + ehdr_task->e_phoff);

	if (phdr_module->p_type != PT_LOAD || phdr_task->p_type != PT_LOAD ||
		phdr_module->p_vaddr != MARS_WORKLOAD_MODULE_BASE_ADDR ||
		phdr_task->p_vaddr != MARS_TASK_BASE_ADDR)
		return MARS_ERROR_FORMAT;

	/* begin process to add the task to the workload queue */
	ret = mars_workload_queue_add_begin(mars, &workload_id, &workload_ea);
	if (ret != MARS_SUCCESS)
		return ret;

	/* prepare work area for task context */
	task = mars_ea_work_area_get(workload_ea,
		MARS_TASK_CONTEXT_ALIGN, MARS_TASK_CONTEXT_SIZE);

	/* initialize the module elf parameters */
	task->module.exec_size = phdr_module->p_filesz;
	task->module.exec_ea =
		mars_ea_map((void *)ehdr_module + phdr_module->p_offset,
			task->module.exec_size);
	if (!task->module.exec_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_map_module;
	}
	task->module.bss_size = phdr_module->p_memsz - phdr_module->p_filesz;
	task->module.entry = ehdr_module->e_entry;

	/* initialize the task elf parameters */
	task->exec_size = phdr_task->p_filesz;
	task->exec_ea =
		mars_ea_map((void *)ehdr_task + phdr_task->p_offset,
			    task->exec_size);
	if (!task->exec_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_map_exec;
	}
	task->bss_size = phdr_task->p_memsz - phdr_task->p_filesz;
	task->entry = ehdr_task->e_entry;

	/* initialize task id */
	task->id.mars_context_ea = mars_ptr_to_ea(mars);
	task->id.workload_id = workload_id;
	if (name)
		strcpy((char *)task->id.name, name);
	else
		task->id.name[0] = 0;

	/* no context save - run complete */
	if (!context_save_unit) {
		task->context_save_unit_ea = 0;
		task->context_save_area_ea = 0;
		goto done;
	}

	/* calculate save unit count and save area size */
	while (p->size) {
		/* check for valid addr/size alignment and high addr boundary */
		if (p->addr & MARS_TASK_CONTEXT_SAVE_UNIT_ADDR_ALIGN_MASK ||
			p->size & MARS_TASK_CONTEXT_SAVE_UNIT_SIZE_ALIGN_MASK ||
			p->addr + p->size > MARS_TASK_CONTEXT_SAVE_SIZE_MAX) {
			ret = MARS_ERROR_PARAMS;
			goto error_context_save_unit_addr_align;
		}

		/* increment save area size and check size limit */
		context_save_area_size += p->size;
		if (context_save_area_size > MARS_TASK_CONTEXT_SAVE_SIZE_MAX) {
			ret = MARS_ERROR_PARAMS;
			goto error_context_save_area_size_limit;
		}

		/* increment save unit count and check count limit */
		context_save_unit_count++;
		if (context_save_unit_count == MARS_TASK_CONTEXT_SAVE_UNIT_MAX)
			break;

		/* increment save unit pointer */
		p++;
	}

	/* allocate context save unit storage */
	task->context_save_unit_ea = mars_ea_memalign(
		MARS_TASK_CONTEXT_SAVE_UNIT_ALIGN,
		context_save_unit_count * MARS_TASK_CONTEXT_SAVE_UNIT_SIZE);
	if (!task->context_save_unit_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_context_save_unit;
	}

	/* allocate context save area */
	task->context_save_area_ea = mars_ea_memalign(
		MARS_TASK_CONTEXT_SAVE_ALIGN, context_save_area_size);
	if (!task->context_save_area_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_context_save_area;
	}

	/* copy the context save struct into allocated storage */
	mars_ea_put(task->context_save_unit_ea,
		context_save_unit,
		context_save_unit_count * MARS_TASK_CONTEXT_SAVE_UNIT_SIZE);

done:
	/* update task context on EA */
	mars_ea_put(workload_ea, task, MARS_TASK_CONTEXT_SIZE);
	mars_ea_sync();

	/* end process to add the task to the workload queue */
	ret = mars_workload_queue_add_end(mars, workload_id);
	if (ret != MARS_SUCCESS)
		goto error_workload_queue_add_end;

	/* return id to caller */
	*id_ret = task->id;

	return MARS_SUCCESS;

error_workload_queue_add_end:
	mars_ea_free(task->context_save_area_ea);
error_malloc_context_save_area:
	mars_ea_free(task->context_save_unit_ea);
error_malloc_context_save_unit:
error_context_save_area_size_limit:
error_context_save_unit_addr_align:
	mars_ea_unmap(task->exec_ea, task->exec_size);
error_map_exec:
	mars_ea_unmap(task->module.exec_ea, task->module.exec_size);
error_map_module:
	mars_workload_queue_add_cancel(mars, workload_id);

	return ret;
}

int mars_task_destroy(struct mars_task_id *id)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	uint64_t workload_ea;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* begin process to remove the task from the workload queue */
	ret = mars_workload_queue_remove_begin(mars, id->workload_id,
		&workload_ea);
	if (ret != MARS_SUCCESS)
		return ret;

	/* prepare work area for task context */
	task = mars_ea_work_area_get(workload_ea,
		MARS_TASK_CONTEXT_ALIGN, MARS_TASK_CONTEXT_SIZE);

	/* get task context from EA */
	mars_ea_get(workload_ea, task, MARS_TASK_CONTEXT_SIZE);

	/* free the allocated context save area if it has one */
	if (task->context_save_area_ea)
		mars_ea_free(task->context_save_area_ea);

	/* free the allocated context save unit storage if it has one */
	if (task->context_save_unit_ea)
		mars_ea_free(task->context_save_unit_ea);

	/* unmap ELFs */
	mars_ea_unmap(task->module.exec_ea, task->module.exec_size);
	mars_ea_unmap(task->exec_ea, task->exec_size);

	/* end process to remove the task from the workload queue */
	ret = mars_workload_queue_remove_end(mars, id->workload_id);
	if (ret != MARS_SUCCESS)
		return ret;

	return MARS_SUCCESS;
}

int mars_task_schedule(struct mars_task_id *id, struct mars_task_args *args,
			uint8_t priority)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	uint64_t workload_ea;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* begin process to schedule the workload in the workload queue */
	ret = mars_workload_queue_schedule_begin(mars, id->workload_id,
		priority, &workload_ea);
	if (ret != MARS_SUCCESS)
		return ret;

	/* prepare work area for task context */
	task = mars_ea_work_area_get(workload_ea,
		MARS_TASK_CONTEXT_ALIGN, MARS_TASK_CONTEXT_SIZE);

	/* get task context from EA */
	mars_ea_get(workload_ea, task, MARS_TASK_CONTEXT_SIZE);

	/* initialize task specific context variables */
	task->stack = 0;
	if (args)
		memcpy(&task->args, args, sizeof(struct mars_task_args));

	/* update task context on EA */
	mars_ea_put(workload_ea, task, MARS_TASK_CONTEXT_SIZE);
	mars_ea_sync();

	/* end process to schedule the workload in the workload queue */
	ret = mars_workload_queue_schedule_end(mars, id->workload_id);
	if (ret != MARS_SUCCESS)
		return ret;

	return MARS_SUCCESS;
}

int mars_task_wait(struct mars_task_id *id, int32_t *exit_code)
{
	int ret;
	struct mars_context *mars;
	uint64_t workload_ea;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* blocking wait for workload completion */
	ret = mars_workload_queue_wait(mars, id->workload_id, &workload_ea);
	if (ret != MARS_SUCCESS)
		return ret;

	/* exit_code requested so return it to caller */
	if (exit_code)
		*exit_code = mars_ea_get_uint32(task_exit_code_ea(workload_ea));

	return MARS_SUCCESS;
}

int mars_task_try_wait(struct mars_task_id *id, int32_t *exit_code)
{
	int ret;
	struct mars_context *mars;
	uint64_t workload_ea;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* non-blocking wait for workload completion */
	ret = mars_workload_queue_try_wait(mars, id->workload_id, &workload_ea);
	if (ret != MARS_SUCCESS)
		return ret;

	/* exit_code requested so return it to caller */
	if (exit_code)
		*exit_code = mars_ea_get_uint32(task_exit_code_ea(workload_ea));

	return MARS_SUCCESS;
}

uint32_t mars_task_get_ticks(void)
{
	return mars_get_ticks();
}
