/*
 * Copyright (C) 2008 Sony Computer Entertainment Inc.
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <mars/mars.h>
#include <libspe2.h>
#include <mars_test.h>
#include <unistd.h>
#include "cmd.h"

#define FAST_DMA_ALIGN	128
#define NUM_TASK	32

#define TASK_STATE_SLEEPING	0
#define TASK_STATE_RUNNING	1

extern spe_program_handle_t mpu_prog;

static struct mars_context *mars_ctx;
static struct mars_task_id task_id[NUM_TASK];
static struct mars_task_args task_args;

struct aligned_int {
	int32_t val;
	int8_t pad[FAST_DMA_ALIGN - sizeof(int32_t)];
};

struct aligned_int cmd[NUM_TASK] __attribute__((aligned(FAST_DMA_ALIGN)));
struct aligned_int res[NUM_TASK] __attribute__((aligned(FAST_DMA_ALIGN)));

int is_running(int task)
{
	res[task].val = 0;
	cmd[task].val = CMD_HELLO;

	/* wait enough time to get reply from the MARS task */
	usleep(100);

	cmd[task].val = CMD_NOP;
	return res[task].val;
}

int main(void)
{
	int i, ret, num_mpu, state[NUM_TASK], yield_target,
		running_task, yielded_task;
	int32_t exit_code;

	num_mpu = spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, -1);
	MARS_TEST_ASSERT(num_mpu > 0);
	MARS_TEST_ASSERT(num_mpu < NUM_TASK);

	ret = mars_context_create(&mars_ctx, 0, 0);
	MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);

	for (i = 0; i < NUM_TASK; i++) {
		ret = mars_task_create(mars_ctx, &task_id[i], NULL,
			mpu_prog.elf_image, MARS_TASK_CONTEXT_SAVE_ALL);
		MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);

		cmd[i].val = CMD_NOP;
		task_args.type.u64[0] = mars_ptr_to_ea(&cmd[i].val);
		task_args.type.u64[1] = mars_ptr_to_ea(&res[i].val);
		ret = mars_task_schedule(&task_id[i], &task_args, 0);
		MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);
	}

	/* Check status of all tasks */
	running_task = 0;
	yield_target = 0;
	for (i = 0; i < NUM_TASK; i++) {
		if (is_running(i)) {
			state[i] = TASK_STATE_RUNNING;
			running_task++;
			yield_target = i;
		} else {
			state[i] = TASK_STATE_SLEEPING;
		}
	}
	MARS_TEST_ASSERT_EQUAL(running_task, num_mpu);

	/* Command one running task to yield */
	cmd[yield_target].val = CMD_YIELD;

	/* sleep enough time to MARS task switch */
	sleep(1);

	/* Check status of all tasks again */
	yielded_task = 0;
	for (i = 0; i < NUM_TASK; i++) {
		if (is_running(i) && (state[i] == TASK_STATE_SLEEPING))
			yielded_task++;
	}

	/* The task issued yield must be sleeping and
	   only one task which have been sleeping must be running */
	MARS_TEST_ASSERT(!is_running(yield_target));
	MARS_TEST_ASSERT_EQUAL(yielded_task, 1);

	/* Command all tasks to exit */
	for (i = 0; i < NUM_TASK; i++)
		cmd[i].val = CMD_EXIT;

	/* Join all tasks */
	for (i = 0; i < NUM_TASK; i++) {
		ret = mars_task_wait(&task_id[i], &exit_code);
		MARS_TEST_ASSERT_EQUAL(ret, 0);
		MARS_TEST_ASSERT_EQUAL(exit_code, 0);

		ret = mars_task_destroy(&task_id[i]);
		MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);
	}

	ret = mars_context_destroy(mars_ctx);
	MARS_TEST_ASSERT_EQUAL(ret, MARS_SUCCESS);

	return 0;
}

